/**************************************************************************
 File name:     interpreter.c  Verson 1.0
 Date:			4/9/07
 Class:         CS480 Principles of Compiler Design
                CS481 Compiler Laboratory
 Assignment:    Pacific University Compiler Project
 Purpose:       Provides the runtime environment for the CS480/481 compiler

 Modifications:
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "../header/pccdefines.h"

void intReadQuads(FILE *pFile);
void intReadStack(FILE *pFile);
int  intDecode(int wMode, int wAddress);
void intExecute(int wOpcode, int wOperand1, int wOperand2, int wOperand3);
void intDumpStack();

int gStack[MAX_RUNTIME];    /* Global runtime stack */
int gSP = 0;                /* Global stack pointer */
int gAP = 0;                /* Global activation record pointer */
int gPC = 0;                /* Global program counter */
int gLevel = 1;             /* Global runtime level */

struct Operand
{
  int wMode;                /* Addressing mode for the operand */
  int wAddress;             /* Address containing operand value */
};

typedef struct Operand OPERAND;
struct structQuad
{
  int gLevel;               /* runtime level of the quad */
  int wOpcode;              /* opcode of the existing quad */
  OPERAND OP1;              /* operand 1 value if it exists */
  OPERAND OP2;              /* operand 2 value if it exists */
  OPERAND OP3;              /* operand 3 value if it exists */
} gQuads[MAX_QUADS];        /* Global runtime Quad array */

/************************************************************************/
/*                                                                      */
/*                              MAIN                                    */
/*                                                                      */
/*  This program interprets the quads generated by the compiler.  It    */
/*  also provides trace facilities for the user.                        */
/*                                                                      */
/*  The machine being simulated is based on a runtime stack.  Each      */
/*  entry in the stack is an integer.  The runtime stack has the        */
/*  following structure:                                                */
/*                                                                      */
/*    location 0          - latest function result                      */
/*    next n locations    - global variables and all constants          */
/*    remaining locations - function activation records. They have the  */
/*                          following form:                             */
/*                                                                      */
/*                          arguments                                   */
/*                          number of arguments                         */
/*                          return address                              */
/*                          old ap                                      */
/*                          local variables                             */
/*                          temporaries                                 */
/*                                                                      */
/*                                                                      */
/*  VARIABLE DEFINITIONS:                                               */
/*                                                                      */
/*   gStack     - the runtime stack                                     */
/*   gQuads     - the quad array                                        */
/*   breakpoint - the opcode at which the user wants to resume tracing  */
/*   trace      - a flag used to indicate if the user wants to trace    */
/*   gAP        - the argument pointer - points to the word containing  */
/*                the old argument pointer in the topmost stack         */
/*                activation record                                     */
/*   gSP        - the stack pointer - points to topmost item on stack   */
/*   gPC        - the program counter - points to next quad to execute  */
/*                                                                      */
/*  NOTES:                                                              */
/*                                                                      */
/*   0 - the format to execute this program is:                         */
/*                                                                      */
/*           interpreter quadfilename [anychar]                         */
/*                                                                      */
/*       The user must supply the name of the quadfile to execute.      */
/*       The filename can be followed with any character to indicate    */
/*       a trace is to occur.                                           */
/*                                                                      */
/*   1 - the quadfile is assumed to have the following form:            */
/*                                                                      */
/*        line 1 to the line beginning with a PROGEND opcode - one      */
/*            quad (7 integers) per line                                */
/*                                                                      */
/*        next line - n, the number of initialized stack words          */
/*                                                                      */
/*        next n lines - initial contents of those words - one per line */
/*                                                                      */
/*   2 - read the quads into the quad array.  Initialize the pc to the  */
/*       location of the PROGBEGIN quad.                                */
/*                                                                      */
/*   3 - initialize the runtime stack.  Initialize the stack pointer    */
/*       to the topmost initialized location.                           */
/*                                                                      */
/*   4 - continue execution until you have reached the PROGEND quad     */
/*                                                                      */
/*   5 - skip over all quads on levels > the current level.             */
/*                                                                      */
/*   5.5 - this function call is the call to main, so we can now        */
/*       increment the level to allow local quads to be executed.       */
/*                                                                      */
/*   6 - a breakpoint of NOP indicates single stepping                  */
/*                                                                      */
/*   7 - if trace and this is the breakpoint, output some trace info.   */
/*                                                                      */
/*   8 - decode the three operands                                      */
/*                                                                      */
/*   9 - if trace and this is the breakpoint, output some more trace    */
/*       info. and ask the user for his desired option.                 */
/*                                                                      */
/*  10 - execute the quad                                               */
/*                                                                      */
/*                                                                      */
/************************************************************************/

int main(int argc, char* argv[])
{
  FILE   *pFile,*fopen();       /* pointer to file for reading quads */
  int    wBreakpoint = OP_NOP;  /* used to indicate single stepping */
  int    fTrace;                /* used to output trace information */
  int    wOpcode = OP_NOP;      /* opcode value of the current quad */
  int    wOperand1;             /* operand 1 value of current quad */
  int    wOperand2;             /* operand 2 value of current quad */
  int    wOperand3;             /* operand 3 value of current quad */

  if (argc == 1)                                                 /* 0 */
  {
    printf("ERROR - must supply quadfile name\n");
    exit(1);
  }

  fTrace = (argc > 2);

  if ((pFile = fopen(argv[1],"r")) == NULL)                      /* 1 */
  {
    printf("ERROR - can't open %s\n",argv[1]);
    exit(1);
  };

  intReadQuads(pFile);                                           /* 2 */

  intReadStack(pFile);                                           /* 3 */

  while (wOpcode != OP_PROGEND)                                  /* 4 */
  {
    while (gQuads[gPC++].gLevel > gLevel); 	                     /* 5 */

    wOpcode = gQuads[--gPC].wOpcode;


    if ((gLevel == 1) && (wOpcode == OP_FUNCCALL))	             /* 5.5 */
      gLevel++;

    if (fTrace && (wBreakpoint == OP_NOP))                       /* 6 */
      wBreakpoint = wOpcode;

    if (fTrace && (wOpcode == wBreakpoint))                      /* 7 */
    {
      printf("\n______________________________________________");
      printf("__________________________________\n");
      printf("TRACE: sp = %d, pc = %d, ap = %d\n",gSP,gPC,gAP);
      printf("TRACE: quad = %d %d %d %d %d %d %d\n",
                                        wOpcode,
                                        gQuads[gPC].OP1.wMode,
                                        gQuads[gPC].OP1.wAddress,
                                        gQuads[gPC].OP2.wMode,
                                        gQuads[gPC].OP2.wAddress,
                                        gQuads[gPC].OP3.wMode,
                                        gQuads[gPC].OP3.wAddress);
     };


     wOperand1 = intDecode(gQuads[gPC].OP1.wMode,                /* 8 */
                           gQuads[gPC].OP1.wAddress);
     wOperand2 = intDecode(gQuads[gPC].OP2.wMode,
                           gQuads[gPC].OP2.wAddress);
     wOperand3 = intDecode(gQuads[gPC].OP3.wMode,
                           gQuads[gPC].OP3.wAddress);


     if (fTrace && (wOpcode == wBreakpoint))                     /* 9 */
     {
       printf("TRACE: operand1 = %d, operand2 = %d, operand3 = %d\n",
                                        wOperand1,wOperand2,wOperand3);
       printf("TRACE: enter [s = stackdump b = breakpoint <cr> = step]: ");

       switch (getchar())
       {
         case 's':
                   intDumpStack();
                   getchar();
                   wBreakpoint = OP_NOP;
                   break;

         case 'b':
                   printf("TRACE: enter breakpoint opcode: ");
                   scanf("%d",&wBreakpoint);
                   getchar();
                   break;

         case '\n':
                   wBreakpoint = OP_NOP;
                   break;

         default:
                   printf("TRACE: terminating execution\n");
                   exit(1);
       };
     };

     intExecute(wOpcode,wOperand1,wOperand2,wOperand3);           /* 10 */
   };

   fclose(pFile);
   printf("\nNormal Termination\n");

   return (0);
}

/************************************************************************/
/*                                                                      */
/*                           intReadQuads                               */
/*                                                                      */
/*  This function reads quads from the quadfile into the quad array.    */
/*                                                                      */
/*  NOTES:                                                              */
/*                                                                      */
/*   1 - the pc must be initialized to the PROGBEGIN quad.              */
/*                                                                      */
/*   2 - the PROGEND quad is physically the last quad in the file       */
/*                                                                      */
/************************************************************************/

void intReadQuads(FILE *pFile)
{
  int  i = -1;

  do
  {
    fscanf(pFile,"%d", &gQuads[++i].gLevel);
    fscanf(pFile,"%d %d %d %d %d %d %d",
                       &gQuads[i].wOpcode,
                       &gQuads[i].OP1.wMode,
                       &gQuads[i].OP1.wAddress,
                       &gQuads[i].OP2.wMode,
                       &gQuads[i].OP2.wAddress,
                       &gQuads[i].OP3.wMode,
                       &gQuads[i].OP3.wAddress);
     if (gQuads[i].wOpcode == OP_PROGBEGIN)               /* 1 */
       gPC = i;
   }

   while (gQuads[i].wOpcode != OP_PROGEND);               /* 2 */
}

/************************************************************************/
/*                                                                      */
/*                           intReadStack                               */
/*                                                                      */
/*  This function reads the initial stack contents from the quadfile.   */
/*                                                                      */
/*  NOTES:                                                              */
/*                                                                      */
/*   1 - read the number of initialized stack words.                    */
/*                                                                      */
/*   2 - initialize the stack words.                                    */
/*                                                                      */
/*   3 - clear the rest of the stack.                                   */
/*                                                                      */
/*   4 - set the stack pointer                                          */
/*                                                                      */
/************************************************************************/

void intReadStack(FILE *pFile)
{
  int  wNumWords;
  int  i;

  fscanf(pFile,"%d",&wNumWords);                               /* 1 */

  for (i = 0;i <= wNumWords - 1;i++)                           /* 2 */
    fscanf(pFile,"%d",&gStack[i]);

  for (i = wNumWords; i < MAX_QUADS; i++)                      /* 3 */
    gStack[i] = 0;

  gSP = wNumWords - 1;                                         /* 4 */
}

/************************************************************************/
/*                                                                      */
/*                               intDecode                              */
/*                                                                      */
/*  This function decodes the passed address based on the passed        */
/*  mode and returns the effective address.  Address modes and the      */
/*  effective addresses yielded by them are as follows:                 */
/*                                                                      */
/*         address mode        |  effective address                     */
/*        _____________________|_____________________________           */
/*         IMMEDIATE           |  address                               */
/*         GLOBAL_LVALUE       |  address                               */
/*         GLOBAL_RVALUE       |  stack(address)                        */
/*         LOCAL_LVALUE        |  gAP + address                         */
/*         LOCAL_RVALUE        |  stack(gAP + address)                  */
/*                                                                      */
/*                                                                      */
/************************************************************************/

int intDecode(int wMode,      /* mode of the operand */
        int wAddress)     /* address of where value for operand is found */
{
  switch(wMode)
  {
    case IMMEDIATE:      return(wAddress);

    case GLOBAL_LVALUE:  return(wAddress);

    case GLOBAL_RVALUE:  return(gStack[wAddress]);

    case LOCAL_LVALUE:   return(gAP + wAddress);

    case LOCAL_RVALUE:   return(gStack[gAP + wAddress]);

  };

  fprintf (stderr, "%s", "Illegal Mode\n");
  exit (1);
}

/************************************************************************/
/*                                                                      */
/*                               VEXECUTE                                */
/*                                                                      */
/*  This function executes the opcode with the operands.  Valid opcodes */
/*  and their actions are as follows:                                   */
/*                                                                      */
/*                                                                      */
/*           ADD             operand3 = operand1 + operand2             */
/*           SUBTRACT        operand3 = operand1 - operand2             */
/*           MULTIPLY        operand3 = operand1 * operand2             */
/*           DIVIDE          operand3 = operand1 / operand2             */
/*           MODULUS         operand3 = operand1 % operand2             */
/*           MINUS           operand3 = -operand1                       */
/*           INCREMENT       operand3 = operand3 + 1                    */
/*           DECREMENT       operand3 = operand3 - 1                    */
/*           DEREFERENCE     operand3 = stack(operand1)                 */
/*           BLT             if (operand1 <  operand2) pc = operand3    */
/*           BGT             if (operand1 >  operand2) pc = operand3    */
/*           BLE             if (operand1 <= operand2) pc = operand3    */
/*           BGE             if (operand1 >= operand2) pc = operand3    */
/*           BNE             if (operand1 != operand2) pc = operand3    */
/*           BEQ             if (operand1 == operand2) pc = operand3    */
/*           BAND            if (operand1 && operand2) pc = operand3    */
/*           BOR             if (operand1 || operand2) pc = operand3    */
/*           BNOT            if (operand1 == 0)        pc = operand3    */
/*           BRA             pc = operand3                              */
/*           PUSHPARAM       push operand1 on stack                     */
/*           FUNCCALL        operand1 = #params already pushed on stack */
/*                           push operand1 on stack                     */
/*                           push pc on stack                           */
/*                           push ap on stack                           */
/*                           ap = sp                                    */
/*                           pc = operand3                              */
/*           FUNCBEGIN       operand1 = size of local data area         */
/*                           sp = sp + operand1                         */
/*           FUNCRETURN      pop activation record off stack            */
/*                           update sp,ap,and pc                        */
/*                           put function result in gStack[0]            */
/*           INPUT           operand1 = # of items to be read           */
/*                           read into the locations specified by the   */
/*                               items on top of the stack.             */
/*           OUTPUT          operand1 = # of items to be written        */
/*                           write the items on top of the stack        */
/*           ASSIGN          operand3 = operand1                        */
/*           PROGBEGIN       place holder                               */
/*           PROGEND         end execution                              */
/*                                                                      */
/************************************************************************/

void intExecute(int wOpcode, /* opcode of current quad to be executed */
           int wOperand1,    /* operand1 value if necessary for opcode */
           int wOperand2,    /* operand2 value if necessary for opcode */
           int wOperand3)    /* operand3 value if necessary for opcode */
{
  int i;

  gPC++;

  switch (wOpcode)
  {
    case OP_ADD:          gStack[wOperand3] = wOperand1 + wOperand2;
                          break;

    case OP_SUBTRACT:     gStack[wOperand3] = wOperand1 - wOperand2;
                          break;

    case OP_MULTIPLY:     gStack[wOperand3] = wOperand1 * wOperand2;
                          break;

    case OP_DIVIDE:       gStack[wOperand3] = wOperand1 / wOperand2;
                          break;

    case OP_MODULUS:      gStack[wOperand3] = wOperand1 % wOperand2;
                          break;

    case OP_MINUS:        gStack[wOperand3] = -wOperand1;
                          break;

    case OP_INCREMENT:    gStack[wOperand3]++;
                          break;

    case OP_DECREMENT:    gStack[wOperand3]--;
                          break;

    case OP_DEREFERENCE:  gStack[wOperand3] = gStack[wOperand1];
                          break;

    case OP_BLT:          if (wOperand1 < wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BGT:          if (wOperand1 > wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BLE:          if (wOperand1 <= wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BGE:          if (wOperand1 >= wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BNE:          if (wOperand1 != wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BEQ:          if (wOperand1 == wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BAND:         if (wOperand1 && wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BOR:          if (wOperand1 || wOperand2)
                            gPC = wOperand3;
                          break;

    case OP_BNOT:         if (wOperand1 == 0)
                            gPC = wOperand3;
                          break;

    case OP_BRA:          gPC = wOperand3;
                          break;

    case OP_PUSHPARAM:    gStack[++gSP] = wOperand1;
                          break;

    case OP_FUNCCALL:     gStack[++gSP] = wOperand1;
                          gStack[++gSP] = gPC;
                          gStack[++gSP] = gAP;
                          gAP = gSP;
                          gPC = wOperand3;
                          break;

    case OP_FUNCBEGIN:    gSP = gSP + wOperand1;
                          break;

    case OP_FUNCRETURN:   gPC = gStack[gAP-1];
                          gSP = (gAP - 3) - (gStack[gAP - 2]);
                          gAP = gStack[gAP];
                          /* reset the level here in case we are called from
                           * the initialization of a global variable
                           */
                          gLevel = gQuads[gPC].gLevel;
                          break;

    case OP_INPUT:        printf("\ninput %d numbers:",wOperand1);
                          for (i = 0;i< wOperand1;i++)
                          {
                            scanf("%d",&gStack[gStack[gSP--]]);
                            scanf(",");
                          }
                          break;

    case OP_OUTPUT:       printf("output: ");
                          for (i = 0;i < wOperand1;i++)
                          {
                            if (i > 0)
                              printf(",");
                            printf("%6d",gStack[gSP--]);
                          };
                          printf("\n");
                          break;

    case OP_ASSIGN:       gStack[wOperand3] = wOperand1;
                          break;

    case OP_PROGBEGIN:    break;

    case OP_PROGEND:      break;
  };
}

/************************************************************************/
/*                                                                      */
/*                           intDumpStack                               */
/*                                                                      */
/*  This function dumps the runtime stack and the stack pointer.        */
/*                                                                      */
/*  NOTES:                                                              */
/*                                                                      */
/*   1 - print contents of stack.                                       */
/*                                                                      */
/*   2 - show where the stack pointer and argument pointer are.         */
/*                                                                      */
/************************************************************************/

void intDumpStack()
{
  int i;
  int j;
  char ch;

  printf("\n_______________________________________________________________");
  printf("\n                  R U N T I M E   S T A C K\n\n");

  for (i = 0;i <= gSP;i = i + 10)                              /* 1 */
  {
    for (j = i;j < i + 10;j++)
    {
      if (j == gSP)					                                  /* 2 */
        ch = 'S';
      else if (j == gAP)
        ch = 'A';
      else
        ch = ' ';

      printf("%5d%c",gStack[j],ch);
    };

    printf("\n");
  };

  printf("\n_______________________________________________________________");
}
